from __future__ import print_function
import msgpackrpc #install as admin: pip install msgpack-rpc-python
import numpy as np #pip install numpy

class MsgpackMixin:

    """
    """
    
    def __repr__(self):
        """
        
        Returns:
            TYPE: Description
        """
        from pprint import pformat
        return "<" + type(self).__name__ + "> " + pformat(vars(self), indent=4, width=1)

    def to_msgpack(self, *args, **kwargs):
        """
        
        Args:
            *args: Description
            **kwargs: Description
        
        Returns:
            TYPE: Description
        """
        return self.__dict__

    @classmethod
    def from_msgpack(cls, encoded):
        """
        
        Args:
            encoded (TYPE): Description
        
        Returns:
            TYPE: Description
        """
        obj = cls()
        #obj.__dict__ = {k.decode('utf-8'): (from_msgpack(v.__class__, v) if hasattr(v, "__dict__") else v) for k, v in encoded.items()}
        obj.__dict__ = { k : (v if not isinstance(v, dict) else getattr(getattr(obj, k).__class__, "from_msgpack")(v)) for k, v in encoded.items()}
        #return cls(**msgpack.unpack(encoded))
        return obj


class ImageType:    

    """
    
    Attributes:
        DepthPerspective (int): Description
        DepthPlanner (int): Description
        DepthVis (int): Description
        DisparityNormalized (int): Description
        Infrared (int): Description
        Scene (int): Description
        Segmentation (int): Description
        SurfaceNormals (int): Description
    """
    
    Scene = 0
    DepthPlanner = 1
    DepthPerspective = 2
    DepthVis = 3
    DisparityNormalized = 4
    Segmentation = 5
    SurfaceNormals = 6
    Infrared = 7

class DrivetrainType:

    """Type of DrivetrainType
    
    Attributes:
        ForwardOnly (int): Fixes yaw along tangent of velocity vector (atan2(vy, vx)) 
        MaxDegreeOfFreedom (int): Description
    """
    
    MaxDegreeOfFreedom = 0
    ForwardOnly = 1
    
class LandedState:

    """
    
    Attributes:
        Flying (int): Description
        Landed (int): Description
    """
    
    Landed = 0
    Flying = 1

class Vector3r(MsgpackMixin):
    """
    
    Attributes:
        x_val (float): Description
        y_val (float): Description
        z_val (float): Description
    """
    
    x_val = 0.0
    y_val = 0.0
    z_val = 0.0

    def __init__(self, x_val = 0.0, y_val = 0.0, z_val = 0.0):
        """
        
        Args:
            x_val (float, optional): Description
            y_val (float, optional): Description
            z_val (float, optional): Description
        """
        self.x_val = x_val
        self.y_val = y_val
        self.z_val = z_val

    @staticmethod
    def nanVector3r():
        """
        
        Returns:
            TYPE: Description
        """
        return Vector3r(np.nan, np.nan, np.nan)

    def __add__(self, other):
        """
        
        Args:
            other (TYPE): Description
        
        Returns:
            TYPE: Description
        """
        return Vector3r(self.x_val + other.x_val, self.y_val + other.y_val, self.z_val + other.z_val)

    def __sub__(self, other):
        """
        
        Args:
            other (TYPE): Description
        
        Returns:
            TYPE: Description
        """
        return Vector3r(self.x_val - other.x_val, self.y_val - other.y_val, self.z_val - other.z_val)

    def __truediv__(self, other):
        """
        
        Args:
            other (TYPE): Description
        
        Returns:
            TYPE: Description
        
        Raises:
            TypeError: Description
        """
        if type(other) in [int, float] + np.sctypes['int'] + np.sctypes['uint'] + np.sctypes['float']:
            return Vector3r( self.x_val / other, self.y_val / other, self.z_val / other)
        else: 
            raise TypeError('unsupported operand type(s) for /: %s and %s' % ( str(type(self)), str(type(other))) )

    def __mul__(self, other):
        """
        
        Args:
            other (TYPE): Description
        
        Returns:
            TYPE: Description
        
        Raises:
            TypeError: Description
        """
        if type(other) in [int, float] + np.sctypes['int'] + np.sctypes['uint'] + np.sctypes['float']:
            return Vector3r(self.x_val*other, self.y_val*other, self.z_val*other)
        else: 
            raise TypeError('unsupported operand type(s) for *: %s and %s' % ( str(type(self)), str(type(other))) )

    def dot(self, other):
        """
        
        Args:
            other (TYPE): Description
        
        Returns:
            TYPE: Description
        
        Raises:
            TypeError: Description
        """
        if type(self) == type(other):
            return self.x_val*other.x_val + self.y_val*other.y_val + self.z_val*other.z_val
        else:
            raise TypeError('unsupported operand type(s) for \'dot\': %s and %s' % ( str(type(self)), str(type(other))) )

    def cross(self, other):
        """
        
        Args:
            other (TYPE): Description
        
        Returns:
            TYPE: Description
        
        Raises:
            TypeError: Description
        """
        if type(self) == type(other):
            cross_product = np.cross(self.to_numpy_array(), other.to_numpy_array())
            return Vector3r(cross_product[0], cross_product[1], cross_product[2])
        else:
            raise TypeError('unsupported operand type(s) for \'cross\': %s and %s' % ( str(type(self)), str(type(other))) )

    def get_length(self):
        """
        
        Returns:
            TYPE: Description
        """
        return ( self.x_val**2 + self.y_val**2 + self.z_val**2 )**0.5

    def distance_to(self, other):
        """
        
        Args:
            other (TYPE): Description
        
        Returns:
            TYPE: Description
        """
        return ( (self.x_val-other.x_val)**2 + (self.y_val-other.y_val)**2 + (self.z_val-other.z_val)**2 )**0.5

    def to_Quaternionr(self):
        """
        
        Returns:
            TYPE: Description
        """
        return Quaternionr(self.x_val, self.y_val, self.z_val, 0)

    def to_numpy_array(self):
        """
        
        Returns:
            TYPE: Description
        """
        return np.array([self.x_val, self.y_val, self.z_val], dtype=np.float32)


class Quaternionr(MsgpackMixin):
    """
    
    Attributes:
        w_val (float): Description
        x_val (float): Description
        y_val (float): Description
        z_val (float): Description
    """
    
    w_val = 0.0
    x_val = 0.0
    y_val = 0.0
    z_val = 0.0

    def __init__(self, x_val = 0.0, y_val = 0.0, z_val = 0.0, w_val = 1.0):
        """
        
        Args:
            x_val (float, optional): Description
            y_val (float, optional): Description
            z_val (float, optional): Description
            w_val (float, optional): Description
        """
        self.x_val = x_val
        self.y_val = y_val
        self.z_val = z_val
        self.w_val = w_val

    @staticmethod
    def nanQuaternionr():
        """
        
        Returns:
            TYPE: Description
        """
        return Quaternionr(np.nan, np.nan, np.nan, np.nan)

    def __add__(self, other):
        """
        
        Args:
            other (TYPE): Description
        
        Returns:
            TYPE: Description
        
        Raises:
            TypeError: Description
        """
        if type(self) == type(other):
            return Quaternionr( self.x_val+other.x_val, self.y_val+other.y_val, self.z_val+other.z_val, self.w_val+other.w_val )
        else:
            raise TypeError('unsupported operand type(s) for +: %s and %s' % ( str(type(self)), str(type(other))) )

    def __mul__(self, other):
        """
        
        Args:
            other (TYPE): Description
        
        Returns:
            TYPE: Description
        
        Raises:
            TypeError: Description
        """
        if type(self) == type(other):
            t, x, y, z = self.w_val, self.x_val, self.y_val, self.z_val
            a, b, c, d = other.w_val, other.x_val, other.y_val, other.z_val
            return Quaternionr( w_val = a*t - b*x - c*y - d*z,
                                x_val = b*t + a*x + d*y - c*z,
                                y_val = c*t + a*y + b*z - d*x,
                                z_val = d*t + z*a + c*x - b*y)
        else:
            raise TypeError('unsupported operand type(s) for *: %s and %s' % ( str(type(self)), str(type(other))) )

    def __truediv__(self, other): 
        """
        
        Args:
            other (TYPE): Description
        
        Returns:
            TYPE: Description
        
        Raises:
            TypeError: Description
        """
        if type(other) == type(self): 
            return self * other.inverse()
        elif type(other) in [int, float] + np.sctypes['int'] + np.sctypes['uint'] + np.sctypes['float']:
            return Quaternionr( self.x_val / other, self.y_val / other, self.z_val / other, self.w_val / other)
        else: 
            raise TypeError('unsupported operand type(s) for /: %s and %s' % ( str(type(self)), str(type(other))) )

    def dot(self, other):
        """
        
        Args:
            other (TYPE): Description
        
        Returns:
            TYPE: Description
        
        Raises:
            TypeError: Description
        """
        if type(self) == type(other):
            return self.x_val*other.x_val + self.y_val*other.y_val + self.z_val*other.z_val + self.w_val*other.w_val
        else:
            raise TypeError('unsupported operand type(s) for \'dot\': %s and %s' % ( str(type(self)), str(type(other))) )

    def cross(self, other):
        """
        
        Args:
            other (TYPE): Description
        
        Returns:
            TYPE: Description
        
        Raises:
            TypeError: Description
        """
        if type(self) == type(other):
            return (self * other - other * self) / 2
        else:
            raise TypeError('unsupported operand type(s) for \'cross\': %s and %s' % ( str(type(self)), str(type(other))) )

    def outer_product(self, other):
        """
        
        Args:
            other (TYPE): Description
        
        Returns:
            TYPE: Description
        
        Raises:
            TypeError: Description
        """
        if type(self) == type(other):
            return ( self.inverse()*other - other.inverse()*self ) / 2
        else:
            raise TypeError('unsupported operand type(s) for \'outer_product\': %s and %s' % ( str(type(self)), str(type(other))) )

    def rotate(self, other):
        """
        
        Args:
            other (TYPE): Description
        
        Returns:
            TYPE: Description
        
        Raises:
            TypeError: Description
            ValueError: Description
        """
        if type(self) == type(other):
            if other.get_length() == 1:
                return other * self * other.inverse()
            else:
                raise ValueError('length of the other Quaternionr must be 1')
        else:
            raise TypeError('unsupported operand type(s) for \'rotate\': %s and %s' % ( str(type(self)), str(type(other))) )        

    def conjugate(self):
        """
        
        Returns:
            TYPE: Description
        """
        return Quaternionr(-self.x_val, -self.y_val, -self.z_val, self.w_val)

    def star(self):
        """
        
        Returns:
            TYPE: Description
        """
        return self.conjugate()

    def inverse(self):
        """
        
        Returns:
            TYPE: Description
        """
        return self.star() / self.dot(self)

    def sgn(self):
        """
        
        Returns:
            TYPE: Description
        """
        return self/self.get_length()

    def get_length(self):
        """
        
        Returns:
            float: norm of quaternion
        """
        return ( self.x_val**2 + self.y_val**2 + self.z_val**2 + self.w_val**2 )**0.5

    def to_numpy_array(self):
        """
        
        Returns:
            np.array: Description
        """
        return np.array([self.x_val, self.y_val, self.z_val, self.w_val], dtype=np.float32)


class Pose(MsgpackMixin):
    """
    
    Attributes:
        orientation (Quaternionr): Description
        position (Vector3r): Description
    """
    
    position = Vector3r()
    orientation = Quaternionr()

    def __init__(self, position_val = Vector3r(), orientation_val = Quaternionr()):
        """
        
        Args:
            position_val (TYPE, optional): Description
            orientation_val (TYPE, optional): Description
        """
        self.position = position_val
        self.orientation = orientation_val

    @staticmethod
    def nanPose():
        """
        
        Returns:
            TYPE: Description
        """
        return Pose(Vector3r.nanVector3r(), Quaternionr.nanQuaternionr())


class CollisionInfo(MsgpackMixin):
    """
    
    Attributes:
        has_collided (bool): Description
        impact_point (TYPE): Description
        normal (TYPE): Description
        object_id (int): Description
        object_name (str): Description
        penetration_depth (float): Description
        position (TYPE): Description
        time_stamp (float): Description
    """
    
    has_collided = False
    normal = Vector3r()
    impact_point = Vector3r()
    position = Vector3r()
    penetration_depth = 0.0
    time_stamp = 0.0
    object_name = ""
    object_id = -1

class YawMode(MsgpackMixin):
    """
    Struct YawMode with two fields, yaw_or_rate and is_rate. 
    If is_rate field is True then yaw_or_rate field is interpreted as angular velocity in **degrees/sec**,
    which means you want vehicle to rotate continuously around its axis at that angular velocity while moving. 

    If is_rate is False then yaw_or_rate is interpreted as angle in **degrees**, 
    which means you want vehicle to rotate to specific angle (i.e. yaw) and keep that angle while moving.

    When yaw_mode.is_rate == true, the drivetrain parameter shouldn't be set to ForwardOnly because there's a contradiction
    as we're asking for the drone to keep front pointing ahead, but also rotate continuously. 
    However if you have yaw_mode.is_rate = false in ForwardOnly mode then you can do some funky stuff. 
    For example, you can have drone do circles and have yaw_or_rate set to 90 so camera is always pointed to center. 
    In MaxDegreeofFreedom also you can get some funky stuff by setting yaw_mode.is_rate = true and say yaw_mode.yaw_or_rate = 20. 
    This will cause drone to go in its path while rotating which may allow to do 360 scanning.

    In most cases, you just don't want yaw to change which you can do by setting yaw rate of 0. 
    The shorthand for this is airsim.YawMode()

    Attributes:
        is_rate (bool): if True, yaw_or_rate is interpreted as angular velocity in **degrees/sec**,
                        if False, yaw_or_rate is interpreted as angles in **degrees**,
        yaw_or_rate (float): value of desired yaw rate, or desired yaw angle. Interpretation depends upon is_rate
    """
    
    is_rate = True
    yaw_or_rate = 0.0

    def __init__(self, is_rate = True, yaw_or_rate = 0.0):
        """
        
        Args:
            is_rate (bool, optional): Description
            yaw_or_rate (float, optional): Description
        """
        self.is_rate = is_rate
        self.yaw_or_rate = yaw_or_rate

class ImageRequest(MsgpackMixin):
    """
    
    Attributes:
        camera_name (str): Description
        compress (bool): Description
        image_type (TYPE): Description
        pixels_as_float (bool): Description
    """
    
    camera_name = '0'
    image_type = ImageType.Scene
    pixels_as_float = False
    compress = False

    def __init__(self, camera_name, image_type, pixels_as_float = False, compress = True):
        """
        
        Args:
            camera_name (TYPE): Description
            image_type (TYPE): Description
            pixels_as_float (bool, optional): Description
            compress (bool, optional): Description
        """
        # todo: in future remove str(), it's only for compatibility to pre v1.2
        self.camera_name = str(camera_name)
        self.image_type = image_type
        self.pixels_as_float = pixels_as_float
        self.compress = compress


class ImageResponse(MsgpackMixin):
    """
    
    Attributes:
        camera_orientation (TYPE): Description
        camera_position (TYPE): Description
        compress (bool): Description
        height (int): Description
        image_data_float (float): Description
        image_data_uint8 (TYPE): Description
        image_type (TYPE): Description
        message (str): Description
        pixels_as_float (float): Description
        time_stamp (TYPE): Description
        width (int): Description
    """
    
    image_data_uint8 = np.uint8(0)
    image_data_float = 0.0
    camera_position = Vector3r()
    camera_orientation = Quaternionr()
    time_stamp = np.uint64(0)
    message = ''
    pixels_as_float = 0.0
    compress = True
    width = 0
    height = 0
    image_type = ImageType.Scene

class KinematicsState(MsgpackMixin):
    """
    
    Attributes:
        angular_acceleration (Vector3r): Description
        angular_velocity (Vector3r): Description
        linear_acceleration (Vector3r): Description
        linear_velocity (Vector3r): Description
        orientation (Quaternionr): Description
        position (Vector3r): Description
    """
    
    position = Vector3r()
    orientation = Quaternionr()
    linear_velocity = Vector3r()
    angular_velocity = Vector3r()
    linear_acceleration = Vector3r()
    angular_acceleration = Vector3r()


class RCData(MsgpackMixin):

    """
    
    Attributes:
        is_initialized (bool): Description
        is_valid (bool): Description
        pitch (TYPE): Description
        roll (TYPE): Description
        switch1 (TYPE): Description
        switch2 (TYPE): Description
        switch3 (TYPE): Description
        switch4 (TYPE): Description
        switch5 (TYPE): Description
        switch6 (TYPE): Description
        switch7 (TYPE): Description
        switch8 (TYPE): Description
        throttle (TYPE): Description
        timestamp (int): Description
        yaw (TYPE): Description
    """
    
    timestamp = 0
    pitch, roll, throttle, yaw = (0.0,)*4 #init 4 variable to 0.0
    switch1, switch2, switch3, switch4 = (0,)*4
    switch5, switch6, switch7, switch8 = (0,)*4
    is_initialized = False
    is_valid = False
    def __init__(self, timestamp = 0, pitch = 0.0, roll = 0.0, throttle = 0.0, yaw = 0.0, switch1 = 0,
                 switch2 = 0, switch3 = 0, switch4 = 0, switch5 = 0, switch6 = 0, switch7 = 0, switch8 = 0, is_initialized = False, is_valid = False):
        """
        
        Args:
            timestamp (int, optional): Description
            pitch (float, optional): Description
            roll (float, optional): Description
            throttle (float, optional): Description
            yaw (float, optional): Description
            switch1 (int, optional): Description
            switch2 (int, optional): Description
            switch3 (int, optional): Description
            switch4 (int, optional): Description
            switch5 (int, optional): Description
            switch6 (int, optional): Description
            switch7 (int, optional): Description
            switch8 (int, optional): Description
            is_initialized (bool, optional): Description
            is_valid (bool, optional): Description
        """
        self.timestamp = timestamp
        self.pitch = pitch 
        self.roll = roll
        self.throttle = throttle 
        self.yaw = yaw 
        self.switch1 = switch1 
        self.switch2 = switch2 
        self.switch3 = switch3 
        self.switch4 = switch4 
        self.switch5 = switch5
        self.switch6 = switch6 
        self.switch7 = switch7 
        self.switch8 = switch8 
        self.is_initialized = is_initialized
        self.is_valid = is_valid

class GeoPoint(MsgpackMixin):
    """
    
    Attributes:
        altitude (float): Description
        latitude (float): Description
        longitude (float): Description
    """
    
    latitude = 0.0
    longitude = 0.0
    altitude = 0.0

class MultirotorState(MsgpackMixin):
    """
    
    Attributes:
        collision (CollisionInfo): Description
        gps_location (GeoPoint): Description
        kinematics_estimated (KinematicsState): Description
        landed_state (LandedState.Landed): Description
        rc_data (RCData): Description
        timestamp (np.uint64): Description
    """
    
    collision = CollisionInfo();
    kinematics_estimated = KinematicsState()
    gps_location = GeoPoint()
    timestamp = np.uint64(0)
    landed_state = LandedState.Landed
    rc_data = RCData()

class ProjectionMatrix(MsgpackMixin):
    """
    
    Attributes:
        matrix (list): Description
    """
    
    matrix = []

class CameraInfo(MsgpackMixin):
    """
    
    Attributes:
        fov (int): Description
        pose (TYPE): Description
        proj_mat (TYPE): Description
    """
    
    pose = Pose()
    fov = -1
    proj_mat = ProjectionMatrix()

class TrajectoryTrackerGains():
    """
    Struct to contain trajectory tracker gains used by the pure pursuit controller for moveBySpline and moveBySplineVelConstraints
    
    Attributes:
        kp_cross_track (float): P gain for position error measured perpendicular to path tangent, or in the "cross track" direction
        kd_cross_track (float): D gain for position error measured perpendicular to path tangent, or in the "cross track" direction
        kp_vel_cross_track (float): P gain for velocity error measured perpendicular to path tangent, or in the "cross track" direction
        kd_vel_cross_track (float): D gain for velocity error measured perpendicular to path tangent, or in the "cross track" direction
        kp_along_track (float): P gain for position error measured along path tangent
        kd_along_track (float): D gain for position error measured along path tangent
        kp_vel_along_track (float): P gain for velocity error measured along path tangent 
        kd_vel_along_track (float): D gain for velocity error measured along path tangent
        kp_z_track (float): P gain for position error measured along world Z axis
        kd_z_track (float): D gain for position error measured along world Z axis
        kp_vel_z (float): P gain for velocity error measured along world Z axis
        kd_vel_z (float): D gain for velocity error measured along world Z axis
        kp_yaw (float): P gain for yaw error
        kd_yaw (float): D gain for yaw error
    """
    
    def __init__(self,
                kp_cross_track = 7.5, 
                kd_cross_track = 0.0, 
                kp_vel_cross_track = 5.0, 
                kd_vel_cross_track = 0.0, 
                kp_along_track = 0.4, 
                kd_along_track = 0.0, 
                kp_vel_along_track = 0.04, 
                kd_vel_along_track = 0.0, 
                kp_z_track = 2.0, 
                kd_z_track = 0.0, 
                kp_vel_z = 0.4, 
                kd_vel_z = 0.0, 
                kp_yaw = 3.0, 
                kd_yaw = 0.1):
        self.kp_cross_track = kp_cross_track
        self.kd_cross_track = kd_cross_track
        self.kp_vel_cross_track = kp_vel_cross_track
        self.kd_vel_cross_track = kd_vel_cross_track
        self.kp_along_track = kp_along_track
        self.kd_along_track = kd_along_track
        self.kp_vel_along_track = kp_vel_along_track
        self.kd_vel_along_track = kd_vel_along_track
        self.kp_z_track = kp_z_track
        self.kd_z_track = kd_z_track
        self.kp_vel_z = kp_vel_z
        self.kd_vel_z = kd_vel_z
        self.kp_yaw = kp_yaw
        self.kd_yaw = kd_yaw

    def to_list(self):
        """
        Call this before sending to setTrajectoryTrackerGains
        
        Returns:
            list[float]: Description
        """
        return [self.kp_cross_track, self.kd_cross_track, self.kp_vel_cross_track, self.kd_vel_cross_track, 
                self.kp_along_track, self.kd_along_track, self.kp_vel_along_track, self.kd_vel_along_track, 
                self.kp_z_track, self.kd_z_track, self.kp_vel_z, self.kd_vel_z, self.kp_yaw, self.kd_yaw]


"""Summary
"""
import numpy as np
import matplotlib.pyplot as plt
import re
import sys
import pdb


def read_pfm(file):
    """Read a pfm file 
    
    Args:
        file (TYPE): Description
    
    Returns:
        TYPE: Description
    
    Raises:
        Exception: Description
    """
    file = open(file, 'rb')

    color = None
    width = None
    height = None
    scale = None
    endian = None

    header = file.readline().rstrip()
    header = str(bytes.decode(header, encoding='utf-8'))
    if header == 'PF':
        color = True
    elif header == 'Pf':
        color = False
    else:
        raise Exception('Not a PFM file.')

    pattern = r'^(\d+)\s(\d+)\s$'
    temp_str = str(bytes.decode(file.readline(), encoding='utf-8'))
    dim_match = re.match(pattern, temp_str)
    if dim_match:
        width, height = map(int, dim_match.groups())
    else:
        temp_str += str(bytes.decode(file.readline(), encoding='utf-8'))
        dim_match = re.match(pattern, temp_str)
        if dim_match:
            width, height = map(int, dim_match.groups())
        else:
            raise Exception('Malformed PFM header: width, height cannot be found')

    scale = float(file.readline().rstrip())
    if scale < 0: # little-endian
        endian = '<'
        scale = -scale
    else:
        endian = '>' # big-endian

    data = np.fromfile(file, endian + 'f')
    shape = (height, width, 3) if color else (height, width)

    data = np.reshape(data, shape)
    # DEY: I don't know why this was there.
    file.close()
    
    return data, scale


def write_pfm(file, image, scale=1):
    """Write a pfm file 
    
    Args:
        file (TYPE): Description
        image (TYPE): Description
        scale (int, optional): Description
    
    Raises:
        Exception: Description
    """
    file = open(file, 'wb')

    color = None

    if image.dtype.name != 'float32':
        raise Exception('Image dtype must be float32.')

    if len(image.shape) == 3 and image.shape[2] == 3: # color image
        color = True
    elif len(image.shape) == 2 or len(image.shape) == 3 and image.shape[2] == 1: # greyscale
        color = False
    else:
        raise Exception('Image must have H x W x 3, H x W x 1 or H x W dimensions.')

    file.write(bytes('PF\n', 'UTF-8') if color else bytes('Pf\n', 'UTF-8'))
    temp_str = '%d %d\n' % (image.shape[1], image.shape[0])
    file.write(bytes(temp_str, 'UTF-8'))

    endian = image.dtype.byteorder

    if endian == '<' or endian == '=' and sys.byteorder == 'little':
        scale = -scale

    temp_str = '%f\n' % scale
    file.write(bytes(temp_str, 'UTF-8'))

    image.tofile(file)



"""
"""
from __future__ import print_function

from .utils import *
from .types import *

import msgpackrpc
import numpy as np
import msgpack
import time
import math
import logging

class MultirotorClient(object):

    """
    
    Attributes:
        client (TYPE): Description
    """
    
    def __init__(self, ip = "", port = 41451, timeout_value = 3600):
        """
        
        Args:
            ip (str, optional): Description
            port (int, optional): Description
            timeout_value (int, optional): Description
        """
        if (ip == ""):
            ip = "127.0.0.1"
        self.client = msgpackrpc.Client(msgpackrpc.Address(ip, port), timeout = timeout_value, pack_encoding = 'utf-8', unpack_encoding = 'utf-8')
        
    # -----------------------------------  Common vehicle APIs ---------------------------------------------
    def reset(self):
        """Resets the drone
        """
        self.client.call('reset')

    def ping(self):
        """
        Returns:
            TYPE: Description
        """
        return self.client.call('ping')

    def getClientVersion(self):
        """
        Returns:
            TYPE: Description
        """
        return 1 # sync with C++ client

    def getServerVersion(self):
        """
        Returns:
            TYPE: Description
        """
        return self.client.call('getServerVersion')

    def getMinRequiredServerVersion(self):
        """
        Returns:
            TYPE: Description
        """
        return 1 # sync with C++ client

    def getMinRequiredClientVersion(self):
        """
        Returns:
            TYPE: Description
        """
        return self.client.call('getMinRequiredClientVersion')

    def enableApiControl(self, vehicle_name = ''):
        """Enables API control for vehicle with name vehicle_name
        
        Args:
            vehicle_name (str, optional): Name of vehicle to enable API control for
        
        Returns:
            bool: Success
        """
        return self.client.call('enableApiControl', True, vehicle_name)

    def disableApiControl(self, vehicle_name = ''):
        """Disables API control for vehicle with name vehicle_name
        
        Args:
            vehicle_name (str, optional): Name of vehicle to disable API control for
        
        Returns:
            bool: Success
        """
        return self.client.call('enableApiControl', False, vehicle_name)

    def isApiControlEnabled(self, vehicle_name = ''):
        """
        
        Args:
            vehicle_name (str, optional): Name of vehicle
        
        Returns:
            bool: If API is enabled 
        """
        return self.client.call('isApiControlEnabled', vehicle_name)

    def arm(self, vehicle_name = ''):
        """Arms vehicle corresponding to vehicle_name
        
        Args:
            vehicle_name (str): Name of vehicle
        
        Returns:
            bool: Success
        
        """
        return self.client.call('armDisarm', True, vehicle_name)

    def disarm(self, vehicle_name = ''):
        """Disarms vehicle corresponding to vehicle_name.
        
        Args:
            vehicle_name (str): Name of vehicle
        
        Returns:
            bool: Success
        
        """
        return self.client.call('armDisarm', False, vehicle_name)

    def simPause(self, is_paused):
        """Pause sim
        
        Args:
            is_paused (TYPE): Description
        """
        self.client.call('simPause', is_paused)

    def simIsPause(self):
        """
        
        Returns:
            TYPE: Description
        """
        return self.client.call("simIsPaused")

    def simContinueForTime(self, seconds):
        """
        
        Args:
            seconds (TYPE): Description
        """
        self.client.call('simContinueForTime', seconds)

    def confirmConnection(self):
        """
        """
        if self.ping():
            print("Connected!")
        else:
             print("Ping returned false!")
        server_ver = self.getServerVersion()
        client_ver = self.getClientVersion()
        server_min_ver = self.getMinRequiredServerVersion()
        client_min_ver = self.getMinRequiredClientVersion()
    
        ver_info = "Client Ver:" + str(client_ver) + " (Min Req: " + str(client_min_ver) + \
              "), Server Ver:" + str(server_ver) + " (Min Req: " + str(server_min_ver) + ")"

        if server_ver < server_min_ver:
            print(ver_info, file=sys.stderr)
            print("AirSim server is of older version and not supported by this client. Please upgrade!")
        elif client_ver < client_min_ver:
            print(ver_info, file=sys.stderr)
            print("AirSim client is of older version and not supported by this server. Please upgrade!")
        else:
            print(ver_info)
        print('')

    # camera control
    # simGetImage returns compressed png in array of bytes
    # image_type uses one of the ImageType members
    def simGetImage(self, camera_name, image_type, vehicle_name = ''):
        """
        Args:
            camera_name (TYPE): Description
            image_type (TYPE): Description
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            TYPE: Description
        """
        # todo: in future remove below, it's only for compatibility to pre v1.2
        camera_name = str(camera_name)

        # because this method returns std::vector<uint8>, msgpack decides to encode it as a string unfortunately.
        result = self.client.call('simGetImage', camera_name, image_type, vehicle_name)
        if (result == "" or result == "\0"):
            return None
        return result

    # camera control
    # simGetImage returns compressed png in array of bytes
    # image_type uses one of the ImageType members
    def simGetImages(self, requests, vehicle_name = ''):
        """
        Args:
            requests (TYPE): Description
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            TYPE: Description
        """
        responses_raw = self.client.call('simGetImages', requests, vehicle_name)
        return [ImageResponse.from_msgpack(response_raw) for response_raw in responses_raw]

    def simGetCollisionInfo(self, vehicle_name = ''):
        """
        
        Args:
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            TYPE: Description
        """
        return CollisionInfo.from_msgpack(self.client.call('simGetCollisionInfo', vehicle_name))

    def simSetVehiclePose(self, pose, ignore_collison, vehicle_name = ''):
        """
        
        Args:
            pose (TYPE): Description
            ignore_collison (TYPE): Description
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        """
        self.client.call('simSetVehiclePose', pose, ignore_collison, vehicle_name)

    def simGetVehiclePose(self, vehicle_name = ''):
        """
        
        Args:
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            TYPE: Description
        """
        pose = self.client.call('simGetVehiclePose', vehicle_name)
        return Pose.from_msgpack(pose)

    def simGetObjectPose(self, object_name):
        """
        
        Args:
            object_name (TYPE): Description
        
        Returns:
            TYPE: Description
        """
        pose = self.client.call('simGetObjectPose', object_name)
        return Pose.from_msgpack(pose)

    def simSetObjectPose(self, object_name, pose, teleport = True):
        """
        
        Args:
            object_name (TYPE): Description
            pose (TYPE): Description
            teleport (bool, optional): Description
        
        Returns:
            TYPE: Description
        """
        return self.client.call('simSetObjectPose', object_name, pose, teleport)

    def simListSceneObjects(self, name_regex = '.*'):
        """Returns list of all objects in the scene 
        
        Args:
            name_regex (str, optional): Description
        
        Returns:
            List[strings]: List of all objects in the scene 
        """
        return self.client.call('simListSceneObjects', name_regex)

    def simLoadLevel(self, level_name):
        """Loads desired level
        
        Args:
            level_name (str): Description
        
        Returns:
            TYPE: Description
        """
        return self.client.call('simLoadLevel', level_name)

    def simSetSegmentationObjectID(self, mesh_name, object_id, is_name_regex = False):
        """
        
        Args:
            mesh_name (str): Description
            object_id (int): Description
            is_name_regex (bool, optional): Description
        
        Returns:
            bool: Success
        """
        return self.client.call('simSetSegmentationObjectID', mesh_name, object_id, is_name_regex)

    def simGetSegmentationObjectID(self, mesh_name):
        """
        
        Args:
            mesh_name (str): Description
        
        Returns:
            int: segmentation ID
        """
        return self.client.call('simGetSegmentationObjectID', mesh_name)

    def simPrintLogMessage(self, message, message_param = "", severity = 0):
        """
        
        Args:
            message (TYPE): Description
            message_param (str, optional): Description
            severity (int, optional): Description
        
        Returns:
            TYPE: Description
        """
        return self.client.call('simPrintLogMessage', message, message_param, severity)

    def simGetCameraInfo(self, camera_name, vehicle_name = ''):
        """
        
        Args:
            camera_name (TYPE): Description
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            CameraInfo: Description
        """
        # TODO: below str() conversion is only needed for legacy reason and should be removed in future
        return CameraInfo.from_msgpack(self.client.call('simGetCameraInfo', str(camera_name), vehicle_name))

    def simSetCameraOrientation(self, camera_name, orientation, vehicle_name = ''):
        """Sets camera's orientation to desired quaternion
        
        Args:
            camera_name (str): Description
            orientation (Quaternionr): Desired orientation quaternion
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        """
        # TODO: below str() conversion is only needed for legacy reason and should be removed in future
        self.client.call('simSetCameraOrientation', str(camera_name), orientation, vehicle_name)

    def simGetGroundTruthKinematics(self, vehicle_name = ''):
        """
        
        Args:
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            TYPE: Description
        """
        kinematics_state = self.client.call('simGetGroundTruthKinematics', vehicle_name)
        return KinematicsState.from_msgpack(kinematics_state)
    simGetGroundTruthKinematics.__annotations__ = {'return': KinematicsState}

    def cancelLastTask():
        """
        """
        self.client.call('cancelLastTask')

    def waitOnLastTask(timeout_sec = float('nan')):
        """
        
        Args:
            timeout_sec (TYPE, optional): Description
        
        Returns:
            TYPE: Description
        """
        return self.client.call('waitOnLastTask', timeout_sec)

# -----------------------------------  Multirotor APIs ---------------------------------------------

    def takeoffAsync(self, timeout_sec = 20, vehicle_name = ''):
        """
        
        Args:
            timeout_sec (int, optional): Description
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            bool: Success
        """
        return self.client.call_async('takeoff', timeout_sec, vehicle_name)  

    def landAsync(self, timeout_sec = 60, vehicle_name = ''):
        """
        
        Args:
            timeout_sec (int, optional): Description
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            bool: Success
        """
        return self.client.call_async('land', timeout_sec, vehicle_name)   

    def goHomeAsync(self, timeout_sec = 3e+38, vehicle_name = ''):
        """
        
        Args:
            timeout_sec (float, optional): Description
            vehicle_name (str, optional): Name of the multirotor to send this command to 

        Returns:
            TYPE: Description
        """
        return self.client.call_async('goHome', timeout_sec, vehicle_name)

    # APIs for control
    def moveByRollPitchYawZAsync(self, roll, pitch, yaw, z, duration, vehicle_name = ''):
        """
        z is given in local NED frame of the vehicle.  
        Roll (rate) / pitch (rate) / yaw (rate) set soints are given in **radians**, in the body frame. 
        The body frame follows the Front Left Up (FLU) convention, and right-handedness. 

        - X axis is along the "Front" direction of the quadrotor. 
            Clockwise rotation about this axis defines a positive **roll** angle.    
            Hence, rolling with a positive angle is equivalent to translating in the "right" direction, w.r.t. our FLU body frame. 
        - Y axis is along the "Left" direction of the quadrotor.  
            Clockwise rotation about this axis defines a positive **pitch** angle.    
            Hence, pitching with a positive angle is equivalent to translating in the "front" direction, w.r.t. our FLU body frame. 
        - Z axis is along the "Up" direction. 
            Clockwise rotation about this axis defines a positive **yaw** angle. 
            Hence, yawing with a positive angle is equivalent to rotated towards the "left" direction wrt our FLU body frame. Or in an anticlockwise fashion in the body XY / FL plane. 
        
        Args:
            roll (float): Desired roll angle, in radians.
            pitch (float): Desired pitch angle, in radians.
            yaw (float): Desired yaw angle, in radians.
            z (float): Desired Z value (in local NED frame of the vehicle)
            duration (float): Desired amount of time (seconds), to send this command for
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            bool: Success
        """
        return self.client.call_async('moveByRollPitchYawZ', roll, -pitch, -yaw, z, duration, vehicle_name)

    def moveByRollPitchYawThrottleAsync(self, roll, pitch, yaw, throttle, duration, vehicle_name = ''):
        """
        Desired throttle is between 0.0 to 1.0
        Roll (rate) / pitch (rate) / yaw (rate) set soints are given in **radians**, in the body frame. 
        The body frame follows the Front Left Up (FLU) convention, and right-handedness. 

        - X axis is along the "Front" direction of the quadrotor. 
            Clockwise rotation about this axis defines a positive **roll** angle.    
            Hence, rolling with a positive angle is equivalent to translating in the "right" direction, w.r.t. our FLU body frame. 
        - Y axis is along the "Left" direction of the quadrotor.  
            Clockwise rotation about this axis defines a positive **pitch** angle.    
            Hence, pitching with a positive angle is equivalent to translating in the "front" direction, w.r.t. our FLU body frame. 
        - Z axis is along the "Up" direction. 
            Clockwise rotation about this axis defines a positive **yaw** angle. 
            Hence, yawing with a positive angle is equivalent to rotated towards the "left" direction wrt our FLU body frame. Or in an anticlockwise fashion in the body XY / FL plane. 
                
        Args:
            roll (float): Desired roll angle, in radians.
            pitch (float): Desired pitch angle, in radians.
            yaw (float): Desired yaw angle, in radians.
            throttle (float): Desired throttle (between 0.0 to 1.0)
            duration (float): Desired amount of time (seconds), to send this command for
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            bool: Success
        """
        return self.client.call_async('moveByRollPitchYawThrottle', roll, -pitch, -yaw, throttle, duration, vehicle_name)

    def moveByRollPitchYawrateThrottleAsync(self, roll, pitch, yaw_rate, throttle, duration, vehicle_name = ''):
        """
        Desired throttle is between 0.0 to 1.0
        Roll (rate) / pitch (rate) / yaw (rate) set soints are given in **radians**, in the body frame. 
        The body frame follows the Front Left Up (FLU) convention, and right-handedness. 

        - X axis is along the "Front" direction of the quadrotor. 
            Clockwise rotation about this axis defines a positive **roll** angle.    
            Hence, rolling with a positive angle is equivalent to translating in the "right" direction, w.r.t. our FLU body frame. 
        - Y axis is along the "Left" direction of the quadrotor.  
            Clockwise rotation about this axis defines a positive **pitch** angle.    
            Hence, pitching with a positive angle is equivalent to translating in the "front" direction, w.r.t. our FLU body frame. 
        - Z axis is along the "Up" direction. 
            Clockwise rotation about this axis defines a positive **yaw** angle. 
            Hence, yawing with a positive angle is equivalent to rotated towards the "left" direction wrt our FLU body frame. Or in an anticlockwise fashion in the body XY / FL plane. 
                
        Args:
            roll (float): Desired roll angle, in radians.
            pitch (float): Desired pitch angle, in radians.
            yaw_rate (float): Desired yaw rate, in radian per second.
            throttle (float): Desired throttle (between 0.0 to 1.0)
            duration (float): Desired amount of time (seconds), to send this command for
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            bool: Success
        """
        return self.client.call_async('moveByRollPitchYawrateThrottle', roll, -pitch, -yaw_rate, throttle, duration, vehicle_name)

    def moveByRollPitchYawrateZAsync(self, roll, pitch, yaw_rate, z, duration, vehicle_name = ''):
        """
        z is given in local NED frame of the vehicle.  
        Roll (rate) / pitch (rate) / yaw (rate) set soints are given in **radians**, in the body frame. 
        The body frame follows the Front Left Up (FLU) convention, and right-handedness. 

        - X axis is along the "Front" direction of the quadrotor. 
            Clockwise rotation about this axis defines a positive **roll** angle.    
            Hence, rolling with a positive angle is equivalent to translating in the "right" direction, w.r.t. our FLU body frame. 
        - Y axis is along the "Left" direction of the quadrotor.  
            Clockwise rotation about this axis defines a positive **pitch** angle.    
            Hence, pitching with a positive angle is equivalent to translating in the "front" direction, w.r.t. our FLU body frame. 
        - Z axis is along the "Up" direction. 
            Clockwise rotation about this axis defines a positive **yaw** angle. 
            Hence, yawing with a positive angle is equivalent to rotated towards the "left" direction wrt our FLU body frame. Or in an anticlockwise fashion in the body XY / FL plane. 
                
        Args:
            roll (float): Desired roll angle, in radians.
            pitch (float): Desired pitch angle, in radians.
            yaw_rate (float): Desired yaw rate, in radian per second.
            z (float): Desired Z value (in local NED frame of the vehicle)
            duration (float): Desired amount of time (seconds), to send this command for
            vehicle_name (str, optional): Name of the multirotor to send this command to 
       
        Returns:
            bool: Success
        """
        return self.client.call_async('moveByRollPitchYawrateZ', roll, -pitch, -yaw_rate, z, duration, vehicle_name)

    def moveByAngleRatesZAsync(self, roll_rate, pitch_rate, yaw_rate, z, duration, vehicle_name = ''):
        """
        z is given in local NED frame of the vehicle.  
        Roll (rate) / pitch (rate) / yaw (rate) set soints are given in **radians**, in the body frame. 
        The body frame follows the Front Left Up (FLU) convention, and right-handedness. 

        - X axis is along the "Front" direction of the quadrotor. 
            Clockwise rotation about this axis defines a positive **roll** angle.    
            Hence, rolling with a positive angle is equivalent to translating in the "right" direction, w.r.t. our FLU body frame. 
        - Y axis is along the "Left" direction of the quadrotor.  
            Clockwise rotation about this axis defines a positive **pitch** angle.    
            Hence, pitching with a positive angle is equivalent to translating in the "front" direction, w.r.t. our FLU body frame. 
        - Z axis is along the "Up" direction. 
            Clockwise rotation about this axis defines a positive **yaw** angle. 
            Hence, yawing with a positive angle is equivalent to rotated towards the "left" direction wrt our FLU body frame. Or in an anticlockwise fashion in the body XY / FL plane. 
        
        Args:
            roll_rate (float): Desired roll rate, in radians / second
            pitch_rate (float): Desired pitch rate, in radians / second
            yaw_rate (float): Desired yaw rate, in radians / second
            z (float): Desired Z value (in local NED frame of the vehicle)
            duration (float): Desired amount of time (seconds), to send this command for
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            bool: Success
        """
        return self.client.call_async('moveByAngleRatesZ', roll_rate, -pitch_rate, -yaw_rate, z, duration, vehicle_name)

    def moveByAngleRatesThrottleAsync(self, roll_rate, pitch_rate, yaw_rate, throttle, duration, vehicle_name = ''):
        """
        Desired throttle is between 0.0 to 1.0
        Roll (rate) / pitch (rate) / yaw (rate) set soints are given in **radians**, in the body frame. 
        The body frame follows the Front Left Up (FLU) convention, and right-handedness. 

        - X axis is along the "Front" direction of the quadrotor. 
            Clockwise rotation about this axis defines a positive **roll** angle.    
            Hence, rolling with a positive angle is equivalent to translating in the "right" direction, w.r.t. our FLU body frame. 
        - Y axis is along the "Left" direction of the quadrotor.  
            Clockwise rotation about this axis defines a positive **pitch** angle.    
            Hence, pitching with a positive angle is equivalent to translating in the "front" direction, w.r.t. our FLU body frame. 
        - Z axis is along the "Up" direction. 
            Clockwise rotation about this axis defines a positive **yaw** angle. 
            Hence, yawing with a positive angle is equivalent to rotated towards the "left" direction wrt our FLU body frame. Or in an anticlockwise fashion in the body XY / FL plane. 
        
        Args:
            roll_rate (float): Desired roll rate, in radians / second
            pitch_rate (float): Desired pitch rate, in radians / second
            yaw_rate (float): Desired yaw rate, in radians / second
            throttle (float): Desired throttle (between 0.0 to 1.0)
            duration (float): Desired amount of time (seconds), to send this command for
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            bool: Success
        """
        return self.client.call_async('moveByAngleRatesThrottle', roll_rate, -pitch_rate, -yaw_rate, throttle, duration, vehicle_name)

    def moveByVelocityAsync(self, vx, vy, vz, duration, drivetrain = DrivetrainType.MaxDegreeOfFreedom, yaw_mode = YawMode(), vehicle_name = ''):
        """
        
        Args:
            vx (float): Description
            vy (float): Description
            vz (float): Description
            duration (float): Description
            drivetrain (TYPE, optional): Description
            yaw_mode (TYPE, optional): Description
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            bool: Success
        """
        return self.client.call_async('moveByVelocity', vx, vy, vz, duration, drivetrain, yaw_mode, vehicle_name)

    def moveByVelocityZAsync(self, vx, vy, z, duration, drivetrain = DrivetrainType.MaxDegreeOfFreedom, yaw_mode = YawMode(), vehicle_name = ''):
        """
        
        Args:
            vx (TYPE): Description
            vy (TYPE): Description
            z (TYPE): Description
            duration (TYPE): Description
            drivetrain (DrivetrainType, optional): Description
            yaw_mode (YawMode, optional): Description
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            bool: Success
        """
        return self.client.call_async('moveByVelocityZ', vx, vy, z, duration, drivetrain, yaw_mode, vehicle_name)

    def moveOnPathAsync(self, path, velocity, timeout_sec = 3e+38, drivetrain = DrivetrainType.MaxDegreeOfFreedom, yaw_mode = YawMode(), 
        lookahead = -1, adaptive_lookahead = 1, vehicle_name = ''):
        """
        
        Args:
            path (TYPE): Description
            velocity (TYPE): Description
            timeout_sec (float, optional): Description
            drivetrain (TYPE, optional): Description
            yaw_mode (TYPE, optional): Description
            lookahead (TYPE, optional): Description
            adaptive_lookahead (int, optional): Description
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            TYPE: Description
        """
        return self.client.call_async('moveOnPath', path, velocity, timeout_sec, drivetrain, yaw_mode, lookahead, adaptive_lookahead, vehicle_name)

    def setTrajectoryTrackerGains(self, gains, vehicle_name = ''):
        """
        Sets trajectory tracker gains for moveOnSplineAsync, moveOnSplineVelConstraintsAsync. 
        Must be called once before either moveOnSpline* API is called
        
        Args:
            gains (list[float]): Pass TrajectoryTrackerGains().to_list() to set default values. Look at TrajectoryTrackerGains to set custom gains
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        """
        self.client.call('setTrajectoryTrackerGains', gains, vehicle_name)

    def moveOnSplineAsync(self, waypoints, vel_max=15.0, acc_max=7.5, add_curr_odom_position_constraint=True, add_curr_odom_velocity_constraint=True, viz_traj=True, vehicle_name = ''):
        """
        - Fits a minimum jerk trajectory to the list of given 3D waypoints (specified by the waypoints parameter).
        - Uses ETHZ-ASL's [mav_trajectory_generation](https://github.com/ethz-asl/mav_trajectory_generation) as the trajectory planning backend.
        - Tracks the references positions and velocities using a pure pursuit tracking controller. 
        The gains of the pure pursuit tracking controller are set by setTrajectoryTrackerGains.
        - Reference yaws are allocated along the tangent of the trajectory. Hence the drone will always look at the direction along which it is flying, behaving like a 3D car.
        - Note: setTrajectoryTrackerGains() must be called once before calling moveOnSpline()

        Args:
            waypoints (list[Vector3r]): List of 3D waypoints, defined in local NED frame of the vehicle to track.
            vel_max (float, optional): Maximum magnitude of velocity along trajectory
            acc_max (float, optional): Maximum magnitude of acceleration along trajectory
            add_curr_odom_position_constraint (bool, optional): Add a start constraint of current linear velocity vector, so that the planned trajectory is smooth if the drone is already moving. 
            add_curr_odom_velocity_constraint (bool, optional): Should only be set to true if add_curr_odom_position_constraint is True. 
                If this is set to true, the trajectory planning backend will add a velocity vector constraint equivalent to the current odometry / linear velocity, so that the planned trajectory is smooth if the drone is already moving. 
            viz_traj (bool, optional): set this to True. visualize trajectory in unreal window. Note that visualization would appear in the FPV image, so this should only be used for debugging. 
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            bool: Success
        """
        return self.client.call_async('moveOnSpline', waypoints, add_curr_odom_position_constraint, add_curr_odom_velocity_constraint, vel_max, acc_max, viz_traj, vehicle_name)

    def moveOnSplineVelConstraintsAsync(self, waypoints, velocity_constraints, vel_max=15.0, acc_max=7.5, add_curr_odom_position_constraint=True, add_curr_odom_velocity_constraint=True, viz_traj=True, vehicle_name = ''):
        """
        - Fits a minimum jerk trajectory to the list of given 3D waypoints (specified by the waypoints parameter).
        - Also adds corresponding 3D velocity vector constraints (specified by the velocity_constraints parameter).
        - Uses ETHZ-ASL's [mav_trajectory_generation](https://github.com/ethz-asl/mav_trajectory_generation) as the trajectory planning backend.
        - Tracks the references positions and velocities using a pure pursuit tracking controller. 
        The gains of the pure pursuit tracking controller are set by setTrajectoryTrackerGains.
        - Reference yaws are allocated along the tangent of the trajectory. Hence the drone will always look at the direction along which it is flying, behaving like a 3D car.
        - Reference yaws are allocated along the tangent of the trajectory. 
        - Note: setTrajectoryTrackerGains() must be called once before calling moveOnSpline()
        
        Args:
            waypoints (list[Vector3r]): List of 3D waypoints, defined in local NED frame of the vehicle to track.
            velocity_constraints (list[Vector3r]): List of 3D velocity vector constraints, defined in local NED frame of the vehicle to track.
            vel_max (float, optional): Maximum magnitude of velocity along trajectory
            acc_max (float, optional): Maximum magnitude of acceleration along trajectory
            add_curr_odom_position_constraint (bool, optional): Add a start constraint of current linear velocity vector, so that the planned trajectory is smooth if the drone is already moving. 
            add_curr_odom_velocity_constraint (bool, optional): Should only be set to true if add_curr_odom_position_constraint is True. 
                If this is set to true, the trajectory planning backend will add a velocity vector constraint equivalent to the current odometry / linear velocity, so that the planned trajectory is smooth if the drone is already moving. 
            viz_traj (bool, optional): set this to True. visualize trajectory in unreal window. Note that visualization would appear in the FPV image, so this should only be used for debugging. 
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            bool: Success
        """
        return self.client.call_async('moveOnSplineVelConstraints', waypoints, velocity_constraints, add_curr_odom_position_constraint, add_curr_odom_velocity_constraint, vel_max, acc_max, viz_traj, vehicle_name)

    def moveToPositionAsync(self, x, y, z, velocity, timeout_sec = 3e+38, drivetrain = DrivetrainType.MaxDegreeOfFreedom, yaw_mode = YawMode(), 
        lookahead = -1, adaptive_lookahead = 1, vehicle_name = ''):
        """
        
        Args:
            x (float): Description
            y (float): Description
            z (float): Description
            velocity (float): Description
            timeout_sec (float, optional): Description
            drivetrain (DrivetrainType, optional): Description
            yaw_mode (YawMode, optional): Description
            lookahead (float, optional): Description
            adaptive_lookahead (int, optional): Description
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            bool: Success
        """
        return self.client.call_async('moveToPosition', x, y, z, velocity, timeout_sec, drivetrain, yaw_mode, lookahead, adaptive_lookahead, vehicle_name)

    def moveToZAsync(self, z, velocity, timeout_sec = 3e+38, yaw_mode = YawMode(), lookahead = -1, adaptive_lookahead = 1, vehicle_name = ''):
        """
        
        Args:
            z (float): Description
            velocity (float): Description
            timeout_sec (float, optional): Description
            yaw_mode (YawMode, optional): Description
            lookahead (float, optional): Description
            adaptive_lookahead (int, optional): Description
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            bool: Success
        """
        return self.client.call_async('moveToZ', z, velocity, timeout_sec, yaw_mode, lookahead, adaptive_lookahead, vehicle_name)

    def moveToYawAsync(self, yaw, timeout_sec = 3e+38, margin = 5, vehicle_name = ''):
        """
        
        Args:
            yaw_rate (float): Desired yaw angle, in **degrees per second**.
            timeout_sec (float, optional): Description
            margin (int, optional): Description
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            bool: Success
        """
        return self.client.call_async('rotateToYaw', yaw, timeout_sec, margin, vehicle_name)

    def moveByYawRateAsync(self, yaw_rate, duration, vehicle_name = ''):
        """
        
        Args:
            yaw_rate (float): Desired yaw rate, in **degrees per second**.
            duration (float): Description
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            bool: Success
        """
        return self.client.call_async('rotateByYawRate', yaw_rate, duration, vehicle_name)

    def hoverAsync(self, vehicle_name = ''):
        """
        
        Args:
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            bool: Success
        """
        return self.client.call_async('hover', vehicle_name)

    def plot_transform(self, pose_list, vehicle_name = ''):
        """
        Plots a transform for a requested list of poses

        Args:
            pose_list (list[Pose]): Description
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        """
        self.client.call('plot_tf', pose_list, 10.0, vehicle_name) 
      
    # query vehicle state
    def getMultirotorState(self, vehicle_name = ''):
        """
        
        Args:
            vehicle_name (str, optional): Name of the multirotor to send this command to 
        
        Returns:
            MultirotorState: Description
        """
        return MultirotorState.from_msgpack(self.client.call('getMultirotorState', vehicle_name))
    getMultirotorState.__annotations__ = {'return': MultirotorState}



import numpy as np #pip install numpy
import math
import time
import sys
import os
import inspect
import types
import re

from .types import *


def string_to_uint8_array(bstr):
    """
    
    Args:
        bstr (TYPE): Description
    
    Returns:
        TYPE: Description
    """
    return np.fromstring(bstr, np.uint8)
    
def string_to_float_array(bstr):
    """
    
    Args:
        bstr (TYPE): Description
    
    Returns:
        TYPE: Description
    """
    return np.fromstring(bstr, np.float32)
    
def list_to_2d_float_array(flst, width, height):
    """
    
    Args:
        flst (TYPE): Description
        width (TYPE): Description
        height (TYPE): Description
    
    Returns:
        TYPE: Description
    """
    return np.reshape(np.asarray(flst, np.float32), (height, width))
    
def get_pfm_array(response):
    """
    
    Args:
        response (TYPE): Description
    
    Returns:
        TYPE: Description
    """
    return list_to_2d_float_array(response.image_data_float, response.width, response.height)

    
def get_public_fields(obj):
    """
    
    Args:
        obj (TYPE): Description
    
    Returns:
        TYPE: Description
    """
    return [attr for attr in dir(obj)
                            if not (attr.startswith("_") 
                            or inspect.isbuiltin(attr)
                            or inspect.isfunction(attr)
                            or inspect.ismethod(attr))]


    
def to_dict(obj):
    """
    
    Args:
        obj (TYPE): Description
    
    Returns:
        TYPE: Description
    """
    return dict([attr, getattr(obj, attr)] for attr in get_public_fields(obj))

    
def to_str(obj):
    """
    
    Args:
        obj (TYPE): Description
    
    Returns:
        TYPE: Description
    """
    return str(to_dict(obj))

    
def write_file(filename, bstr):
    """
    
    Args:
        filename (TYPE): Description
        bstr (TYPE): Description
    """
    with open(filename, 'wb') as afile:
        afile.write(bstr)

# helper method for converting getOrientation to roll/pitch/yaw
# https:#en.wikipedia.org/wiki/Conversion_between_quaternions_and_Euler_angles
    
def to_eularian_angles(q):
    """
    
    Args:
        q (TYPE): Description
    
    Returns:
        TYPE: Description
    """
    z = q.z_val
    y = q.y_val
    x = q.x_val
    w = q.w_val
    ysqr = y * y

    # roll (x-axis rotation)
    t0 = +2.0 * (w*x + y*z)
    t1 = +1.0 - 2.0*(x*x + ysqr)
    roll = math.atan2(t0, t1)

    # pitch (y-axis rotation)
    t2 = +2.0 * (w*y - z*x)
    if (t2 > 1.0):
        t2 = 1
    if (t2 < -1.0):
        t2 = -1.0
    pitch = math.asin(t2)

    # yaw (z-axis rotation)
    t3 = +2.0 * (w*z + x*y)
    t4 = +1.0 - 2.0 * (ysqr + z*z)
    yaw = math.atan2(t3, t4)

    return (pitch, roll, yaw)

    
def to_quaternion(pitch, roll, yaw):
    """
    
    Args:
        pitch (TYPE): Description
        roll (TYPE): Description
        yaw (TYPE): Description
    
    Returns:
        TYPE: Description
    """
    t0 = math.cos(yaw * 0.5)
    t1 = math.sin(yaw * 0.5)
    t2 = math.cos(roll * 0.5)
    t3 = math.sin(roll * 0.5)
    t4 = math.cos(pitch * 0.5)
    t5 = math.sin(pitch * 0.5)

    q = Quaternionr()
    q.w_val = t0 * t2 * t4 + t1 * t3 * t5 #w
    q.x_val = t0 * t3 * t4 - t1 * t2 * t5 #x
    q.y_val = t0 * t2 * t5 + t1 * t3 * t4 #y
    q.z_val = t1 * t2 * t4 - t0 * t3 * t5 #z
    return q

    
def wait_key(message = ''):
    '''Wait for a key press on the console and return it. 
    
    Args:
        message (str, optional): Description
    
    Returns:
        TYPE: Description
    '''
    if message != '':
        print (message)

    result = None
    if os.name == 'nt':
        import msvcrt
        result = msvcrt.getch()
    else:
        import termios
        fd = sys.stdin.fileno()

        oldterm = termios.tcgetattr(fd)
        newattr = termios.tcgetattr(fd)
        newattr[3] = newattr[3] & ~termios.ICANON & ~termios.ECHO
        termios.tcsetattr(fd, termios.TCSANOW, newattr)

        try:
            result = sys.stdin.read(1)
        except IOError:
            pass
        finally:
            termios.tcsetattr(fd, termios.TCSAFLUSH, oldterm)

    return result

    
def read_pfm(file):
    """Read a pfm file 
    
    Args:
        file (TYPE): Description
    
    Returns:
        TYPE: Description
    
    Raises:
        Exception: Description
    """
    file = open(file, 'rb')

    color = None
    width = None
    height = None
    scale = None
    endian = None

    header = file.readline().rstrip()
    header = str(bytes.decode(header, encoding='utf-8'))
    if header == 'PF':
        color = True
    elif header == 'Pf':
        color = False
    else:
        raise Exception('Not a PFM file.')

    temp_str = str(bytes.decode(file.readline(), encoding='utf-8'))
    dim_match = re.match(r'^(\d+)\s(\d+)\s$', temp_str)
    if dim_match:
        width, height = map(int, dim_match.groups())
    else:
        raise Exception('Malformed PFM header.')

    scale = float(file.readline().rstrip())
    if scale < 0: # little-endian
        endian = '<'
        scale = -scale
    else:
        endian = '>' # big-endian

    data = np.fromfile(file, endian + 'f')
    shape = (height, width, 3) if color else (height, width)

    data = np.reshape(data, shape)
    # DEY: I don't know why this was there.
    file.close()
    
    return data, scale

    
def write_pfm(file, image, scale=1):
    """Write a pfm file 
    
    Args:
        file (TYPE): Description
        image (TYPE): Description
        scale (int, optional): Description
    
    Raises:
        Exception: Description
    """
    file = open(file, 'wb')

    color = None

    if image.dtype.name != 'float32':
        raise Exception('Image dtype must be float32.')

    if len(image.shape) == 3 and image.shape[2] == 3: # color image
        color = True
    elif len(image.shape) == 2 or len(image.shape) == 3 and image.shape[2] == 1: # grayscale
        color = False
    else:
        raise Exception('Image must have H x W x 3, H x W x 1 or H x W dimensions.')

    file.write('PF\n'.encode('utf-8')  if color else 'Pf\n'.encode('utf-8'))
    temp_str = '%d %d\n' % (image.shape[1], image.shape[0])
    file.write(temp_str.encode('utf-8'))

    endian = image.dtype.byteorder

    if endian == '<' or endian == '=' and sys.byteorder == 'little':
        scale = -scale

    temp_str = '%f\n' % scale
    file.write(temp_str.encode('utf-8'))

    image.tofile(file)

    
def write_png(filename, image):
    """image must be numpy array H X W X channels
    
    Args:
        filename (TYPE): Description
        image (TYPE): Description
    """
    import zlib, struct

    buf = image.flatten().tobytes()
    width = image.shape[1]
    height = image.shape[0]

    # reverse the vertical line order and add null bytes at the start
    width_byte_3 = width * 3
    raw_data = b''.join(b'\x00' + buf[span:span + width_byte_3]
                        for span in range((height - 1) * width_byte_3, -1, - width_byte_3))

    def png_pack(png_tag, data):
        """
        
        Args:
            png_tag (TYPE): Description
            data (TYPE): Description
        
        Returns:
            TYPE: Description
        """
        chunk_head = png_tag + data
        return (struct.pack("!I", len(data)) +
                chunk_head +
                struct.pack("!I", 0xFFFFFFFF & zlib.crc32(chunk_head)))

    png_bytes = b''.join([
        b'\x89PNG\r\n\x1a\n',
        png_pack(b'IHDR', struct.pack("!2I5B", width, height, 8, 6, 0, 0, 0)),
        png_pack(b'IDAT', zlib.compress(raw_data, 9)),
        png_pack(b'IEND', b'')])

    write_file(filename, png_bytes)
